"""
This module provides the SNMP security model for community based v1 exchanges
"""
from typing import Any, Dict
from warnings import warn

from x690.types import Integer, OctetString, Sequence

from puresnmp.credentials import V1, Credentials
from puresnmp.exc import ErrorResponse, SnmpError
from puresnmp.pdu import PDU
from puresnmp.plugins.security import SecurityModel

IDENTIFIER = 1


class SNMPv1SecurityModel(SecurityModel[PDU, Sequence]):
    """
    Implementation of the SNMPv1 community based security model
    """

    def generate_request_message(
        self,
        message: PDU,
        security_engine_id: bytes,  # pylint: disable=unused-argument
        credentials: Credentials,
    ) -> Sequence:
        if not isinstance(credentials, V1):
            raise SnmpError(
                "Credentials for the SNMPv1 security model must be "
                "V1 instances!"
            )

        # The SNMPv1 support is blatantly ripping off the SNMPv2c
        # implementation which is *mostly* equivalent. Once the distinctions
        # have been ironed out, this warning can be removed.
        warn("Experimental SNMPv1 support", UserWarning)

        packet = Sequence(
            [Integer(0), OctetString(credentials.community), message]
        )
        return packet

    def process_incoming_message(
        self,
        message: Sequence,
        credentials: Credentials,
    ) -> PDU:
        proto_version, community, pdu = message
        if not isinstance(credentials, V1):
            raise SnmpError(
                "Credentials for the SNMPv1 security model must be "
                "V1 instances!"
            )

        # The SNMPv1 support is blatantly ripping off the SNMPv2c
        # implementation which is *mostly* equivalent. Once the distinctions
        # have been ironed out, this warning can be removed.
        warn("Experimental SNMPv1 support", UserWarning)

        if proto_version.pythonize() != 0:
            raise SnmpError(
                "Incoming SNMP message is not supported by the SNMPv1 "
                "security model. Most likely the device is not talking "
                f"SNMPv1 but rather a different SNMP version ({proto_version})."
            )

        if community.pythonize() != credentials.community.encode("ascii"):
            raise SnmpError("Mismatching community in response message!")

        return pdu  # type: ignore


def create(local_config: Dict[bytes, Dict[str, Any]]) -> SNMPv1SecurityModel:
    """
    Create a new instance of the security model
    """
    return SNMPv1SecurityModel(local_config)



from typing import Any, NamedTuple, Optional

from x690.types import Null, ObjectIdentifier, X690Type


class VarBind(NamedTuple):
    """
    A "VarBind" is a 2-tuple containing an object-identifier and the
    corresponding value.
    """

    oid: ObjectIdentifier = ObjectIdentifier()
    value: X690Type[Any] = Null()


class PyVarBind(NamedTuple):
    """
    A "PyVarBind" is a 2-tuple containing an object-identifier and the
    corresponding value using native Python data-types
    """

    oid: str = ""
    value: Optional[Any] = None

    @staticmethod
    def from_raw(raw_varbind: VarBind) -> "PyVarBind":
        return PyVarBind(
            raw_varbind.oid.pythonize(), raw_varbind.value.pythonize()
        )



"""
This module provides the high-level wrapper :py:class:`.PyWrapper` around
:py:class:`puresnmp.api.raw.Client`.

Care is taken to make this as pythonic as possible and hide as many of the gory
implementations as possible.

This module provides "syntactic sugar" around the lower-level, but almost
identical, module :py:mod:`puresnmp.api.raw`.

While this "pythonic" API returns native Python types, the "raw" module
returns the variable types unmodified which are all subclasses of
:py:class:`x690.types.X690Type`.


>>> import asyncio
>>> from puresnmp import Client, V2C, PyWrapper
>>>
>>> async def example():
...    client = PyWrapper(Client("192.0.2.1", V2C("public")))
...    output = await client.get("1.3.6.1.2.1.1.1.0")
...    return output
"""

import logging
from collections import OrderedDict
from datetime import timedelta
from typing import Any, AsyncGenerator, Dict, List

from x690.types import ObjectIdentifier, X690Type

from ..const import ERRORS_STRICT
from ..pdu import Trap
from ..util import BulkResult, TTableRow
from ..varbind import PyVarBind
from . import raw

LOG = logging.getLogger(__name__)
TWalkResponse = AsyncGenerator[PyVarBind, None]


class PyWrapper:
    """
    A wrapper around a :py:class:`puresnmp.api.raw.Client` instance.

    The wrapper ensures conversion of internal API data-types to and from
    Python-native types.

    Using Python native types shields from internal changes internally in
    :py:mod:`puresnmp` at the cost of flexibility. Most applications should
    mostly benefit from this.

    In cases internal data-types are still wanted, one can access the
    ``.client`` attribute of PyWrapper instances which exposes the same API but
    with internally used data-types.
    """

    client: raw.Client

    def __init__(self, client: raw.Client) -> None:
        self.client = client

    async def get(self, oid: str) -> Any:
        """
        Delegates to :py:meth:`puresnmp.api.raw.Client.get` but
        converts internal types to simple Python types.
        """
        oid_internal = ObjectIdentifier(oid)
        raw_value = await self.client.get(oid_internal)
        return raw_value.pythonize()

    async def getnext(self, oid: str) -> PyVarBind:
        """
        Delegates to :py:meth:`puresnmp.api.raw.Client.getnext` but
        converts internal types to simple Python types.
        """
        varbind = await self.client.getnext(ObjectIdentifier(oid))
        return PyVarBind.from_raw(varbind)

    async def set(self, oid: str, value: X690Type[Any]) -> Dict[str, Any]:
        """
        Delegates to :py:meth:`puresnmp.api.raw.Client.set` but
        converts internal types to simple Python types.
        """
        result = await self.multiset({oid: value})
        return result[oid.lstrip(".")]  # type: ignore

    async def multiset(
        self, mappings: Dict[str, X690Type[Any]]
    ) -> Dict[str, Any]:
        """
        Delegates to :py:meth:`puresnmp.api.raw.Client.multiset` but
        converts internal types to simple Python types.
        """

        mappings_internal = {
            ObjectIdentifier(oid): value for oid, value in mappings.items()
        }
        raw_output = await self.client.multiset(mappings_internal)
        pythonized = {
            str(oid): value.pythonize() for oid, value in raw_output.items()
        }
        return pythonized

    async def walk(
        self,
        oid: str,
        errors: str = ERRORS_STRICT,
    ) -> TWalkResponse:
        """
        Delegates to :py:meth:`puresnmp.api.raw.Client.walk` but
        converts internal types to simple Python types.
        """
        raw_result = self.client.walk(ObjectIdentifier(oid), errors)
        async for varbind in raw_result:
            yield PyVarBind.from_raw(varbind)

    async def multiwalk(self, oids: List[str]) -> TWalkResponse:
        """
        Delegates to :py:meth:`puresnmp.api.raw.Client.multiwalk` but
        converts internal types to simple Python types.
        """
        oids_internal = [ObjectIdentifier(oid) for oid in oids]
        async for varbind in self.client.multiwalk(oids_internal):
            yield PyVarBind.from_raw(varbind)

    async def multiget(self, oids: List[str]) -> List[Any]:
        """
        Delegates to :py:meth:`puresnmp.api.raw.Client.multiget` but
        converts internal types to simple Python types.
        """
        oids_internal = [ObjectIdentifier(oid) for oid in oids]
        raw_output = await self.client.multiget(oids_internal)
        pythonized = [value.pythonize() for value in raw_output]
        return pythonized

    async def bulkwalk(
        self,
        oids: List[str],
        bulk_size: int = 10,
    ) -> TWalkResponse:
        """
        Delegates to :py:meth:`puresnmp.api.raw.Client.bulkwalk` but
        converts internal types to simple Python types.
        """
        oids_internal = [ObjectIdentifier(oid) for oid in oids]
        result = self.client.bulkwalk(oids_internal, bulk_size=bulk_size)
        async for varbind in result:
            yield PyVarBind.from_raw(varbind)

    async def bulkget(
        self,
        scalar_oids: List[str],
        repeating_oids: List[str],
        max_list_size: int = 10,
    ) -> BulkResult:
        """
        Delegates to :py:meth:`puresnmp.api.raw.Client.bulkget` but
        converts internal types to simple Python types.
        """

        scalar_oids_int = [ObjectIdentifier(oid) for oid in scalar_oids]
        repeating_oids_int = [ObjectIdentifier(oid) for oid in repeating_oids]
        raw_output = await self.client.bulkget(
            scalar_oids_int,
            repeating_oids_int,
            max_list_size=max_list_size,
        )
        pythonized_scalars = {
            oid: value.pythonize() for oid, value in raw_output.scalars.items()
        }
        pythonized_list = OrderedDict(
            [
                (oid, value.pythonize())
                for oid, value in raw_output.listing.items()
            ]
        )
        return BulkResult(pythonized_scalars, pythonized_list)

    async def table(
        self,
        oid: str,
        _rowtype: TTableRow = Dict[str, Any],  # type: ignore
    ) -> List[Dict[str, Any]]:
        """
        Delegates to :py:meth:`puresnmp.api.raw.Client.table` but
        converts internal types to simple Python types.
        """
        tmp: List[TTableRow] = await self.client.table(
            ObjectIdentifier(oid), _rowtype=_rowtype
        )
        output = []
        for row in tmp:
            index = row.pop("0")
            pythonized = {key: value.pythonize() for key, value in row.items()}
            pythonized["0"] = index
            output.append(pythonized)
        return output

    async def bulktable(
        self,
        oid: str,
        bulk_size: int = 10,
        _rowtype: TTableRow = Dict[str, Any],  # type: ignore
    ) -> List[TTableRow]:
        """
        Delegates to :py:meth:`puresnmp.api.raw.Client.bulktable` but
        converts internal types to simple Python types.
        """
        tmp: List[TTableRow] = await self.client.bulktable(
            ObjectIdentifier(oid), bulk_size=bulk_size, _rowtype=_rowtype
        )
        output: List[TTableRow] = []
        for row in tmp:
            index = row.pop("0")
            pythonized = {key: value.pythonize() for key, value in row.items()}
            pythonized["0"] = index
            output.append(pythonized)  # type: ignore
        return output


class TrapInfo:
    """
    This class wraps a :py:class:`puresnmp.pdu.Trap` instance (accessible via
    ``raw_trap`` and makes values available as "pythonic" values.
    """

    #: The raw Trap PDU.
    #:
    #: .. warning::
    #:     This will leak data-types which are used internally by ``puresnmp``
    #:     and may change even in minor updates. You should, if possible use
    #:     the values from the properties on this object.  This exist mainly to
    #:     expose values which can be helpful in debugging.  If something is
    #:     missing from the properties, please open a corresponding support
    #:     ticket!
    raw_trap: Trap

    def __init__(self, raw_trap: Trap) -> None:
        self.raw_trap = raw_trap

    def __repr__(self):
        # type: () -> str
        return "<TrapInfo from %s on %s with %d values>" % (
            self.origin,
            self.oid,
            len(self.values),
        )

    @property
    def origin(self) -> str:
        """
        Accesses the IP-Address from which the trap was sent

        May be the empty string if the source is unknown
        """
        if self.raw_trap is None or self.raw_trap.source is None:
            return ""
        return self.raw_trap.source.address

    @property
    def uptime(self) -> timedelta:
        """
        Returns the uptime of the device.
        """
        return self.raw_trap.value.varbinds[0].value.pythonize()  # type: ignore

    @property
    def oid(self) -> str:
        """
        Returns the Trap-OID
        """
        return self.raw_trap.value.varbinds[1].value.pythonize()  # type: ignore

    @property
    def values(self):
        # type: () -> Dict[str, Any]
        """
        Returns all the values contained in this trap as dictionary mapping
        OIDs to values.
        """
        output = {}
        for varbind in self.raw_trap.value.varbinds[2:]:
            pyvarbind = PyVarBind.from_raw(varbind)
            output[pyvarbind.oid] = pyvarbind.value
        return output



"""
This module contains helper-classes for the different credentials used in SNMP.

These credentials also define the underlying message-processing model by
their very nature. V1 credentials will use the V1 message processing model,
V2C uses the community-based V2 message-processing model and so on.
"""
from typing import NamedTuple, Optional


class Auth(NamedTuple):
    """
    Configuration for "authentication" values in SNMPv3
    """

    #: The authentication key (password)
    key: bytes
    #: The authentication method (md5, sha, ...)
    method: str


class Priv(NamedTuple):
    """
    Configuration for "privacy/encryption" values in SNMPv3
    """

    #: The privacy key (password)
    key: bytes
    #: The privacy method (des, aes, ...)
    method: str


class Credentials:
    """
    Parent class for SNMP credentials. This should not be used other than
    type-hinting. Use the concrete classes for SNMP calls.
    """

    __slots__ = ["mpm"]

    #: The message-processing model identifier
    mpm: int

    def __init__(self, mpm: int) -> None:
        self.mpm = mpm

    def __repr__(self) -> str:
        return f"<{__name__}.{self.__class__.__name__}>"


class V1(Credentials):
    """
    Credentials for SNMPv1 exchanges
    """

    __slots__ = ["community"]

    #: The community-string
    community: str

    def __init__(self, community: str) -> None:
        super().__init__(0)
        self.community = community

    def __eq__(self, other: object) -> bool:
        return isinstance(other, V1) and other.community == self.community


class V2C(V1):
    """
    Credentials for community-based SNMPv2 exchanges
    """

    #: The community-string
    community: str

    def __init__(self, community: str) -> None:
        super().__init__(community)
        self.mpm = 1


class V3(Credentials):
    """
    Credentials for SNMPv3 exchanges
    """

    __slots__ = ["username", "auth", "priv"]

    #: The local username for SNMP exchanges
    username: str

    #: Authentication details. If left at "None", authentication is disabled.
    auth: Optional[Auth]

    #: Encryption details. If left at "None", encryption is disabled. Note
    #: that for encryption to be enabled, authentication must also be enabled.
    priv: Optional[Priv]

    def __init__(
        self,
        username: str,
        auth: Optional[Auth] = None,
        priv: Optional[Priv] = None,
    ) -> None:
        super().__init__(3)
        self.username = username
        self.auth = auth
        self.priv = priv

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, V3)
            and other.username == self.username
            and other.auth == self.auth
            and other.priv == self.priv
        )



"""
This module provides a plugin architecture for security methods.

Each security plugin can be distributed as separate package by providing
modules inside the namespace-package "puresnmp_plugins.security". Note that
in order to be a valid namespace-package, such a package *must not* have a
``__init__.py`` file!

Example folder-structure for a privacy plugin:

.. code-block:: text

    my-security-plugin/
     +- setup.py (or pyproject.toml)
     +- puresnmp/
         +- security/
             +- mymodule.py
             +- myothermodule.py

Note that there is no ``__init__.py`` file!

In order for modules to be detected as plugin, they must follow the following
rules:

* Have a no-arg function ``create`` returning a instance of the class (or
  subclass) of :py:class:`puresnmp.security.SecurityModel`.
* Contain a int-variable ``IDENTIFIER``. This variable should map to the
  security model identifiers as defined in the SNMPv3 standard. See
  https://www.iana.org/assignments/snmp-number-spaces/snmp-number-spaces.xhtml
  and :rfc:`3411`
"""
from types import ModuleType
from typing import Any, Awaitable, Callable, Dict, Generic, Optional, TypeVar
from warnings import warn

from puresnmp.credentials import Credentials
from puresnmp.exc import UnknownSecurityModel
from puresnmp.plugins.pluginbase import Loader

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore

#: The type of an *unsecured* message inside of "puresnmp"
TPureSNMPType = TypeVar("TPureSNMPType", bound=Any)

#: The type of a *secured* message outside of "puresnmp"
TX690Type = TypeVar("TX690Type", bound=Any)


class TSecurityPlugin(Protocol):
    """
    Protocol for security plugins
    """

    # pylint: disable=too-few-public-methods

    def create(
        self, local_config: Dict[bytes, Dict[str, Any]]
    ) -> "SecurityModel[TPureSNMPType, TX690Type]":  # pragma: no cover
        """
        Create a new instance of a security model
        """
        ...


class SecurityModel(Generic[TPureSNMPType, TX690Type]):
    """
    Each Security Model defines the applied protecion on SNMP PDUs
    """

    #: The "Local Configuration Datastore" (LCD). This contains contextual
    #: information which may be needed for some security models (as defined by
    #: the SNMPv3 architecture.)
    local_config: Dict[bytes, Dict[str, Any]]

    def __init__(self, local_config: Dict[bytes, Dict[str, Any]]) -> None:
        self.local_config = local_config

    def generate_request_message(
        self,
        message: TPureSNMPType,
        security_engine_id: bytes,
        credentials: Credentials,
    ) -> TX690Type:  # pragma: no cover
        """
        Take a plain unprocessed message and applies security to the message
        as defined by the concrete security model.

        It returns the processed message including security modifications.
        Further processing is provided by the "message-processing-model" to
        prepare it for sending out to the network.

        The kind of processing applied to the message depends on the
        credential type.

        :param message: The original message
        :param security_engine_id: The engine-id of the receiving SNMP engine
        :param credentials: A credential object giving information on what
            kind of operations are needed on the message.
        """
        raise NotImplementedError(f"Not yet implemented in {self.__class__}")

    def process_incoming_message(
        self,
        message: TX690Type,
        credentials: Credentials,
    ) -> TPureSNMPType:  # pragma: no cover
        """
        Takes a message which included potential security modifications (like
        encryption) and "undoes" these modifications in order to make the
        message usable again.

        Returns an unprocessed message.

        The kind of processing applied to the message depends on the
        credential type.

        :param message: The original message
        :param credentials: A credential object giving information on what
            kind of operations are needed on the message.
        """
        raise NotImplementedError(f"Not yet implemented in {self.__class__}")

    def set_engine_timing(
        self,
        engine_id: bytes,
        engine_boots: int,
        engine_time: int,
    ) -> None:  # pragma: no cover
        """
        Update the security model with timing information of the remote-engine.

        The timing information is required by SNMPv3 to authenticat/encrypt
        messages.
        """
        raise NotImplementedError(f"Not yet implemented in {self.__class__}")

    async def send_discovery_message(
        self,
        transport_handler: Callable[[bytes], Awaitable[bytes]],
    ) -> Any:  # pragma: no cover
        """
        Send a discovery message to the remote engine.

        :param handler: A callable that is responsible to send data to the
            remote engine. This callable should already be aware of how to
            connect. It takes the bytes to be sent to the remote and should
            return the response bytes without any processing.
        """
        raise NotImplementedError(f"Not yet implemented in {self.__class__}")


def is_valid_sec_plugin(mod: ModuleType) -> bool:
    return all(
        [
            hasattr(mod, "create"),
            hasattr(mod, "IDENTIFIER"),
        ]
    )


def create(
    identifier: int, local_config: Optional[Dict[bytes, Dict[str, Any]]] = None
) -> SecurityModel[TPureSNMPType, TX690Type]:
    """
    Return an instance of the given security module by identifier.

    This looks up the module by "IDENTIFIER" as specified in the given plugin.

    If no plugin with the given identifier is found, a *KeyError* is raised
    """
    if local_config is None:
        warn(
            f"Calling {__name__}.create() without local-config is deprecated!",
            DeprecationWarning,
            stacklevel=2,
        )
        local_config = {}
    namespace = "puresnmp_plugins.security"
    loader = Loader(namespace, is_valid_sec_plugin)
    result = loader.create(identifier)
    if not result:
        raise UnknownSecurityModel(
            namespace,
            identifier,
            sorted(loader.discovered_plugins.keys()),
        )
    return result.create(local_config)  # type: ignore
